let fonts = {
    'Adobe Arabic': 'AdobeArabic-Bold.otf',
    'Andale Mono': 'andalemo.ttf',
    'Arial Black': 'ariblk.ttf',
    'Arial Bold Italic': 'arialbi.ttf',
    'Arial Bold': 'arialbd.ttf',
    'Arial Italic': 'ariali.ttf',
    'Arial Unicode MS': 'arialuni.ttf',
    'Arial': 'arial.ttf',
    'Comic Sans MS Bold': 'comicbd.ttf',
    'Comic Sans MS': 'comic.ttf',
    'Courier New Bold Italic': 'courbi.ttf',
    'Courier New Bold': 'courbd.ttf',
    'Courier New Italic': 'couri.ttf',
    'Courier New': 'cour.ttf',
    'DejaVu LGC Sans Mono Bold Oblique': 'DejaVuLGCSansMono-BoldOblique.ttf',
    'DejaVu LGC Sans Mono Bold': 'DejaVuLGCSansMono-Bold.ttf',
    'DejaVu LGC Sans Mono Oblique': 'DejaVuLGCSansMono-Oblique.ttf',
    'DejaVu LGC Sans Mono': 'DejaVuLGCSansMono.ttf',
    'DejaVu Sans Bold Oblique': 'DejaVuSans-BoldOblique.ttf',
    'DejaVu Sans Bold': 'DejaVuSans-Bold.ttf',
    'DejaVu Sans Condensed Bold Oblique': 'DejaVuSansCondensed-BoldOblique.ttf',
    'DejaVu Sans Condensed Bold': 'DejaVuSansCondensed-Bold.ttf',
    'DejaVu Sans Condensed Oblique': 'DejaVuSansCondensed-Oblique.ttf',
    'DejaVu Sans Condensed': 'DejaVuSansCondensed.ttf',
    'DejaVu Sans ExtraLight': 'DejaVuSans-ExtraLight.ttf',
    'DejaVu Sans Mono Bold Oblique': 'DejaVuSansMono-BoldOblique.ttf',
    'DejaVu Sans Mono Bold': 'DejaVuSansMono-Bold.ttf',
    'DejaVu Sans Mono Oblique': 'DejaVuSansMono-Oblique.ttf',
    'DejaVu Sans Mono': 'DejaVuSansMono.ttf',
    'DejaVu Sans Oblique': 'DejaVuSans-Oblique.ttf',
    'DejaVu Sans': 'DejaVuSans.ttf',
    'Gautami': 'arialuni.ttf',
    'Georgia Bold Italic': 'georgiaz.ttf',
    'Georgia Bold': 'georgiab.ttf',
    'Georgia Italic': 'georgiai.ttf',
    'Georgia': 'georgia.ttf',
    'Impact': 'impact.ttf',
    'mangal': 'arialuni.ttf',
    'Mangal': 'arialuni.ttf',
    'MeeraInimai': 'arialuni.ttf',
    'Noto Sans Thai': 'NotoSansThai.ttf',
    'Rubik Black Italic': 'Rubik-BlackItalic.ttf',
    'Rubik Black': 'Rubik-Black.ttf',
    'Rubik Bold Italic': 'Rubik-BoldItalic.ttf',
    'Rubik Bold': 'Rubik-Bold.ttf',
    'Rubik Italic': 'Rubik-Italic.ttf',
    'Rubik Light Italic': 'Rubik-LightItalic.ttf',
    'Rubik Light': 'Rubik-Light.ttf',
    'Rubik Medium Italic': 'Rubik-MediumItalic.ttf',
    'Rubik Medium': 'Rubik-Medium.ttf',
    'Rubik': 'Rubik-Regular.ttf',
    'Tahoma': 'tahoma.ttf',
    'Times New Roman Bold Italic': 'timesbi.ttf',
    'Times New Roman Bold': 'timesbd.ttf',
    'Times New Roman Italic': 'timesi.ttf',
    'Times New Roman': 'times.ttf',
    'Trebuchet MS Bold Italic': 'trebucbi.ttf',
    'Trebuchet MS Bold': 'trebucbd.ttf',
    'Trebuchet MS Italic': 'trebucit.ttf',
    'Trebuchet MS': 'trebuc.ttf',
    'Verdana Bold Italic': 'verdanaz.ttf',
    'Verdana Bold': 'verdanab.ttf',
    'Verdana Italic': 'verdanai.ttf',
    'Verdana': 'verdana.ttf',
    'Vrinda Bold': 'arialuni.ttf',
    'Vrinda': 'arialuni.ttf',
    'Webdings': 'webdings.ttf',
}

/**
 * Send message from JS to the app handler.
 * @param type - The type of the message. The current handled types: `LOG`, `JS_LOADED`, `OCTOPUS_ERROR`.
 * @param payload - Any string or object.
 */
function sendMessage(type, payload) {
    let action = { type: type };

    if (payload != null) {
        action.payload = payload;
    }

    try {
        // rendererControllerMessageHandler is declared in SubtitlesRenderer.swift
        webkit.messageHandlers.rendererControllerMessageHandler.postMessage(JSON.stringify(action));
    } catch(e) {
        console.error('the native handler does not exist: ', type);
    }
}

function configureWebPage() {
    document.documentElement.style.webkitUserSelect = 'none';
    document.documentElement.style.webkitTouchCallout = 'none';
}

configureWebPage();

let rendererDebug = false;
let octopus = null;

let canvas = document.getElementById("subtitles-canvas");

function resize() {
    let pixelRatio = window.devicePixelRatio || 1;
    if (octopus != null) {
        octopus.resize(window.outerWidth * pixelRatio, window.outerHeight * pixelRatio, 0, 0);
    }
}

function observeResize() {
    function resizeWithTimeout() {
        setTimeout(resize, 200);
    }

    window.addEventListener('resize', resizeWithTimeout, false);
    window.addEventListener('orientationchange', resizeWithTimeout, false);
    resize();
}

observeResize();

/**
 * Create SubtitlesOctopus instance with either subtitle URL or subtitle content
 * @param url - Network subtitle URL
 * @param content - ASS/SSA subtitle content
 */
function instantiateOctopus(url, content) {
    rendererDispose();

    let subtitleOption = {};
    if (url != null) {
        subtitleOption.subUrl = url;
    } else if (content != null) {
        subtitleOption.subContent = content;
    }

    function octopusOnReady() {
        resize();
    }

    function octopusOnError(error) {
        sendMessage("OCTOPUS_ERROR", error);
    }

    let options = {
        availableFonts: {},
        debug: rendererDebug,
        canvas: canvas,
        fonts: [],
        onReady: octopusOnReady,
        onError: octopusOnError,
        defaultFont: 'arialuni.ttf'
    };
    for (let fontName in fonts) {
        options.availableFonts[fontName.toLowerCase()] = fonts[fontName];
    }

    let mergedOptions = {...options, ...subtitleOption};
    octopus = new SubtitlesOctopus(mergedOptions);
}

function b64DecodeUnicode(content) {
  // Going backwards: from bytestream, to percent-encoding, to original string.
  return decodeURIComponent(atob(content).split('').map(function(c) {
    return '%' + ('00' + c.charCodeAt(0).toString(16)).slice(-2);
  }).join(''));
}

// Public methods

function rendererSetRemoteSubtitleURL(url) {
    if (octopus == null) {
        instantiateOctopus(url, null);
    } else {
        octopus.setTrackByUrl(url);
    }
}

function rendererSetLocalSubtitleContent(stringJSON) {
    const object = JSON.parse(stringJSON);
    const decodedContent = b64DecodeUnicode(object.content);
    if (octopus == null) {
        instantiateOctopus(null, decodedContent);
    } else {
        octopus.setTrack(decodedContent);
    }
}

function rendererSetCurrentTime(seconds) {
    if (octopus != null) {
        octopus.setCurrentTime(seconds);
    }
}

function rendererSetIsPlaying(currentTimeSeconds) {
    if (octopus != null) {
        octopus.setIsPaused(false, currentTimeSeconds);
    }
}

function rendererSetIsPaused(currentTimeSeconds) {
    if (octopus != null) {
        octopus.setIsPaused(true, currentTimeSeconds);
    }
}

function rendererFreeTrack() {
    if (octopus != null) {
        octopus.freeTrack();
    }
}

function rendererDispose() {
    if (octopus != null) {
        octopus.dispose()
        octopus = null;
    }
}

sendMessage("JS_LOADED");
